@@ -20,8 +20,8 @@ from django.forms.models import BaseInlineFormSet
 from django.http import HttpResponseRedirect
 from django.template.response import TemplateResponse
 from django.urls import reverse
+from django.utils.translation import gettext_lazy as _
 from django.utils.translation import ngettext
-from django.utils.translation import ugettext_lazy as _
 from openwisp_utils.admin import UUIDAdmin
 from organizations.base_admin import (
     BaseOrganizationAdmin,
@@ -84,6 +84,7 @@ class RequiredInlineFormSet(BaseInlineFormSet):
 class OrganizationOwnerInline(admin.StackedInline):
     model = OrganizationOwner
     extra = 0
+    raw_id_fields = ('organization_user',)
 
     def has_change_permission(self, request, obj=None):
         if obj and not request.user.is_superuser and not request.user.is_owner(obj):
@@ -95,6 +96,7 @@ class OrganizationUserInline(admin.StackedInline):
     model = OrganizationUser
     formset = RequiredInlineFormSet
     view_on_site = False
+    autocomplete_fields = ('organization',)
 
     def get_formset(self, request, obj=None, **kwargs):
         """
@@ -159,7 +161,7 @@ class UserCreationForm(UserFormMixin, BaseUserCreationForm):
 
     class Meta(BaseUserCreationForm.Meta):
         fields = ['username', 'email', 'password1', 'password2']
-        personal_fields = ['first_name', 'last_name', 'phone_number']
+        personal_fields = ['first_name', 'last_name', 'phone_number', 'birth_date']
         fieldsets = (
             (None, {'classes': ('wide',), 'fields': fields}),
             ('Personal Info', {'classes': ('wide',), 'fields': personal_fields}),
@@ -205,6 +207,7 @@ class UserAdmin(MultitenantAdminMixin, BaseUserAdmin, BaseAdmin):
     inlines = [EmailAddressInline, OrganizationUserInline]
     save_on_top = True
     actions = ['delete_selected_overridden', 'make_inactive', 'make_active']
+    fieldsets = list(BaseUserAdmin.fieldsets)
 
     # To ensure extended apps use this template.
     change_form_template = 'admin/openwisp_users/user/change_form.html'
@@ -315,7 +318,7 @@ class UserAdmin(MultitenantAdminMixin, BaseUserAdmin, BaseAdmin):
         # do not allow operators to escalate their privileges
         if not request.user.is_superuser:
             # copy to avoid modifying reference
-            fields = fields[:] + ['user_permissions']
+            fields = fields[:] + ['user_permissions', 'is_superuser']
         return fields
 
     def has_change_permission(self, request, obj=None):
@@ -406,8 +409,8 @@ class UserAdmin(MultitenantAdminMixin, BaseUserAdmin, BaseAdmin):
 
     def change_view(self, request, object_id, form_url='', extra_context=None):
         extra_context = extra_context or {}
-        obj = self.model.objects.get(pk=object_id)
-        if user_not_allowed_to_change_owner(request.user, obj):
+        obj = self.get_object(request, object_id)
+        if obj is not None and user_not_allowed_to_change_owner(request.user, obj):
             show_owner_warning = True
             extra_context.update({'show_owner_warning': show_owner_warning})
         return super().change_view(request, object_id, form_url, extra_context)
@@ -449,17 +452,46 @@ class UserAdmin(MultitenantAdminMixin, BaseUserAdmin, BaseAdmin):
                 'belong to some organization owners.'
             )
             self.message_user(
-                request,
-                ngettext(single_msg, multiple_msg, not_deleted,),
-                messages.ERROR,
+                request, ngettext(single_msg, multiple_msg, not_deleted), messages.ERROR
             )
         for instance in instances:
             instance.save()
 
 
+class OrganizationUserFilter(admin.SimpleListFilter):
+    """
+    Allows filtering users by the organization they're related to
+    """
+
+    title = _('organization')
+    parameter_name = 'organization'
+
+    def lookups(self, request, model_admin):
+        organizations = Organization.objects.all()
+        if not request.user.is_superuser:
+            organizations = organizations.filter(
+                pk__in=request.user.organizations_managed
+            )
+        lookups = []
+        for org in organizations:
+            lookups.append((str(org.pk), org.name))
+        # show filter only if multiple orgs are accessible
+        if len(lookups) > 1:
+            return lookups
+        return tuple()
+
+    def queryset(self, request, queryset):
+        if self.value():
+            queryset = queryset.filter(
+                openwisp_users_organizationuser__organization=self.value()
+            )
+        return queryset
+
+
 base_fields = list(UserAdmin.fieldsets[1][1]['fields'])
-additional_fields = ['bio', 'url', 'company', 'location', 'phone_number']
+additional_fields = ['bio', 'url', 'company', 'location', 'phone_number', 'birth_date']
 UserAdmin.fieldsets[1][1]['fields'] = base_fields + additional_fields
+UserAdmin.fieldsets.insert(3, ('Internal', {'fields': ('notes',)}))
 UserAdmin.add_fieldsets[0][1]['fields'] = (
     'username',
     'email',
@@ -467,10 +499,14 @@ UserAdmin.add_fieldsets[0][1]['fields'] = (
     'password2',
 )
 UserAdmin.search_fields += ('phone_number',)
+UserAdmin.list_filter = (OrganizationUserFilter,) + UserAdmin.list_filter
 
 
 class GroupAdmin(BaseGroupAdmin, BaseAdmin):
-    pass
+    if 'reversion' in settings.INSTALLED_APPS:
+        # Correctly register the proxy model
+        def reversion_register(self, model, **kwargs):
+            return super().reversion_register(model, for_concrete_model=False, **kwargs)
 
 
 class OrganizationAdmin(
@@ -478,7 +514,7 @@ class OrganizationAdmin(
 ):
     view_on_site = False
     inlines = [OrganizationOwnerInline]
-    readonly_fields = ['uuid']
+    readonly_fields = ['uuid', 'created', 'modified']
     ordering = ['name']
 
     def get_inline_instances(self, request, obj=None):
@@ -585,7 +621,7 @@ class OrganizationUserAdmin(
                 'belong to some organization owners.'
             )
             self.message_user(
-                request, ngettext(single_msg, multiple_msg, count,), messages.ERROR,
+                request, ngettext(single_msg, multiple_msg, count), messages.ERROR
             )
         # otherwise proceed but remove org users from the delete queryset
         return delete_selected(self, request, queryset)
@@ -629,14 +665,11 @@ if allauth_settings.SOCIALACCOUNT_ENABLED:
     ]:
         admin.site.unregister(apps.get_model(*model))
 
-
-if (
-    'rest_framework.authtoken' in settings.INSTALLED_APPS and not settings.DEBUG
-):  # pragma: no cover
-    Token = apps.get_model('authtoken', 'Token')
+if 'rest_framework.authtoken' in settings.INSTALLED_APPS:  # pragma: no cover
+    TokenProxy = apps.get_model('authtoken', 'TokenProxy')
 
     try:
-        admin.site.unregister(Token)
+        admin.site.unregister(TokenProxy)
     except NotRegistered:
         pass
 
