@@ -5,12 +5,10 @@ from openwisp_users.migrations import update_admins_permissions
 
 class Migration(migrations.Migration):
 
-    dependencies = [
-        ('openwisp_users', '0007_unique_email'),
-    ]
+    dependencies = [('openwisp_users', '0007_unique_email')]
 
     operations = [
         migrations.RunPython(
             update_admins_permissions, reverse_code=migrations.RunPython.noop
-        ),
+        )
     ]
