@@ -1,6 +1,7 @@
 import contextlib
-import os
+import re
 import smtplib
+import uuid
 from unittest.mock import patch
 
 from django.contrib import admin
@@ -11,6 +12,7 @@ from django.core.exceptions import ValidationError
 from django.db import DEFAULT_DB_ALIAS
 from django.test import TestCase
 from django.urls import reverse
+from openwisp_utils.tests import capture_any_output
 from swapper import load_model
 
 from ..admin import OrganizationOwnerAdmin
@@ -28,11 +30,9 @@ OrganizationOwner = load_model('openwisp_users', 'OrganizationOwner')
 User = get_user_model()
 Group = load_model('openwisp_users', 'Group')
 
-devnull = open(os.devnull, 'w')
-
 
 class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestCase):
-    """ test admin site """
+    """test admin site"""
 
     app_label = 'openwisp_users'
     is_integration_test = False
@@ -94,6 +94,15 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         self.assertEqual(emailaddress.email, 'test@testadd.com')
         self.assertEqual(len(mail.outbox), 1)
 
+        with self.subTest("Test HTML Template used for Signup Mail"):
+            email = mail.outbox.pop()
+            self.assertTrue(email.alternatives)
+            self.assertIn('<b>testadd</b>', email.alternatives[0][0])
+            self.assertIn(
+                'To confirm this is correct, please click on the button below',
+                email.alternatives[0][0],
+            )
+
     def test_admin_add_user_empty_email(self):
         admin = self._create_admin()
         self.client.force_login(admin)
@@ -116,7 +125,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         params = user.__dict__
         params['email'] = 'new@mail.com'
         params.pop('phone_number')
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
         params = self._additional_params_pop(params)
         # inline emails
@@ -168,6 +178,28 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         self.assertEqual(len(mail.outbox), 0)
         self.assertContains(response, 'errors field-email')
 
+    def test_admin_change_user_page_get_invalid_UUID(self):
+        admin = self._create_admin()
+        self.client.force_login(admin)
+        with self.subTest('Test for wrong identifier'):
+            response = self.client.get(
+                reverse(f'admin:{self.app_label}_user_change', args=['WRONG']),
+                follow=True,
+            )
+            content = 'User with ID “WRONG” doesn’t exist. Perhaps it was deleted?'
+            self.assertContains(response, content, status_code=200)
+        with self.subTest('Test for non-existing user'):
+            id = uuid.uuid4()
+            response = self.client.get(
+                reverse(
+                    f'admin:{self.app_label}_user_change',
+                    args=[id],
+                ),
+                follow=True,
+            )
+            content = f'User with ID “{id}” doesn’t exist. Perhaps it was deleted?'
+            self.assertContains(response, content, status_code=200)
+
     def test_organization_view_on_site(self):
         admin = self._create_admin()
         self.client.force_login(admin)
@@ -235,12 +267,34 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             reverse(f'admin:{self.app_label}_user_change', args=[operator.pk])
         )
         with self.subTest('User Permissions'):
-            html = f'<div class="readonly">{self.app_label}'
-            self.assertContains(response, html)
+            # regex to check if `<div class="readonly"> ... app_label </div>`
+            # exists in the response
+            html = f'<div class="readonly">((?!</div>).)*({self.app_label})'
+            self.assertTrue(
+                re.search(
+                    html,
+                    str(response.content),
+                )
+            )
         with self.subTest('Organization User Inline'):
             html = 'class="readonly"><img src="/static/admin/img/icon'
             self.assertContains(response, html)
 
+    def test_org_manager_privilege_escalation(self):
+        operator = self._create_operator()
+        options = {
+            'organization': self._get_org(),
+            'is_admin': True,
+            'user': self._get_operator(),
+        }
+        self._create_org_user(**options)
+        self.client.force_login(operator)
+        response = self.client.get(
+            reverse(f'admin:{self.app_label}_user_change', args=[operator.pk])
+        )
+        self.assertNotContains(response, 'superuser')
+        self.assertNotContains(response, 'id_user_permissions_from')
+
     def test_admin_changelist_user_superusers_hidden(self):
         self._create_admin()
         operator = self._create_operator()
@@ -282,7 +336,10 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         self.assertContains(response, 'selected>operator-org1</option>')
         self.assertNotContains(response, 'selected>operator-org2</option>')
 
-    def test_operator_can_see_organization_add_user(self):
+    # the autocomplete fields are removed only in this
+    # test to make testing multitenancy simpler
+    @patch('openwisp_users.admin.OrganizationUserInline.autocomplete_fields', [])
+    def test_operator_can_see_organization_add_user(self, *args):
         org1 = self._create_org(name='operator-org1')
         org2 = self._create_org(name='operator-org2')
         operator = self._create_operator()
@@ -435,6 +492,56 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             res, '<li>User with this Email address already exists.</li>'
         )
 
+    def test_create_user_existing_mail_different_case(self):
+        self._create_user(username='user1', email='user@example.com')
+        admin = self._create_admin()
+        self.client.force_login(admin)
+        params = dict(
+            username='testadd',
+            email='USER@example.com',
+            password1='tester',
+            password2='tester',
+        )
+        params.update(self.add_user_inline_params)
+        res = self.client.post(reverse(f'admin:{self.app_label}_user_add'), params)
+        content = '<li>User with this Email address already exists.</li>'
+        self.assertContains(res, content, status_code=200)
+
+    def test_update_user_no_validation_error(self):
+        admin = self._create_admin()
+        self.client.force_login(admin)
+        user = self._create_user(email='user@example.com', username='user1')
+        params = user.__dict__
+        params['username'] = 'user2'
+        params.pop('last_login')
+        params.pop('phone_number')
+        params.pop('password', None)
+        params.pop('_password', None)
+        params = self._additional_params_pop(params)
+        params.update(self.add_user_inline_params)
+        params.update(
+            {
+                'emailaddress_set-TOTAL_FORMS': 1,
+                'emailaddress_set-INITIAL_FORMS': 1,
+                'emailaddress_set-0-verified': True,
+                'emailaddress_set-0-primary': True,
+                'emailaddress_set-0-id': user.emailaddress_set.first().id,
+                'emailaddress_set-0-user': user.id,
+            }
+        )
+        params.update(self._get_user_edit_form_inline_params(user, self._get_org()))
+        res = self.client.post(
+            reverse(f'admin:{self.app_label}_user_change', args=[user.pk]),
+            params,
+            follow=True,
+        )
+        user.refresh_from_db()
+        self.assertNotIn(
+            '<li>User with this Email address already exists.</li>',
+            res.content.decode(),
+        )
+        self.assertEqual(user.username, 'user2')
+
     def test_edit_user_email_exists(self):
         admin = self._create_admin()
         self.client.force_login(admin)
@@ -444,7 +551,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         params = user.__dict__
         params['email'] = 'test@tester.com'
         params.pop('phone_number')
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
         params = self._additional_params_pop(params)
         params.update(self.add_user_inline_params)
@@ -473,7 +581,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         user = self._create_operator()
         self._create_org_user(user=user)
         params = user.__dict__
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
         params.pop('phone_number')
         params.update(self.add_user_inline_params)
@@ -497,7 +606,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         params = user.__dict__
         params['groups'] = str(group.pk)
         params.pop('phone_number')
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
         params.update(self.add_user_inline_params)
         params.update(self._additional_params_add())
@@ -561,7 +671,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
         params['username'] = 'newuser1'
         params['groups'] = str(group.pk)
         params.pop('phone_number')
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
         params.update(self.add_user_inline_params)
         params.update(self._additional_params_add())
@@ -864,7 +975,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             reverse(f'admin:{self.app_label}_organization_change', args=[org.pk])
         )
         self.assertEqual(response.status_code, 200)
-        self.assertContains(response, 'Organization owners')
+        self.assertContains(response, 'Organization owner')
 
     @patch.object(
         OrganizationOwner, 'full_clean', side_effect=ValidationError('invalid')
@@ -1112,7 +1223,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
     def test_can_change_org(self):
         org = self._get_org()
         user = self._create_user(
-            username='change', password='change', email='email@email', is_staff=True
+            username='change', password='change', email='email@email.com', is_staff=True
         )
         group = Group.objects.filter(name='Administrator')
         user.groups.set(group)
@@ -1142,7 +1253,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
 
     def test_only_superuser_has_add_delete_org_perm(self):
         user = self._create_user(
-            username='change', password='change', email='email@email', is_staff=True
+            username='change', password='change', email='email@email.com', is_staff=True
         )
         group = Group.objects.filter(name='Administrator')
         user.groups.set(group)
@@ -1195,10 +1306,10 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
 
     def test_can_change_inline_org_owner(self):
         user1 = self._create_user(
-            username='user1', password='user1', email='email1@email', is_staff=True
+            username='user1', password='user1', email='email1@email.com', is_staff=True
         )
         user2 = self._create_user(
-            username='user2', password='user2', email='email2@email', is_staff=True
+            username='user2', password='user2', email='email2@email.com', is_staff=True
         )
         group = Group.objects.filter(name='Administrator')
         user1.groups.set(group)
@@ -1241,7 +1352,10 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             params.update(self._get_org_edit_form_inline_params(self._get_admin(), org))
             self.client.force_login(self._get_admin())
             user3 = self._create_user(
-                username='user3', password='user3', email='email3@email', is_staff=True
+                username='user3',
+                password='user3',
+                email='email3@email.com',
+                is_staff=True,
             )
             user3.groups.set(group)
             org_user3 = self._create_org_user(
@@ -1256,7 +1370,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
     def test_only_superuser_can_delete_inline_org_owner(self):
         org = self._get_org()
         user = self._create_user(
-            username='change', password='change', email='email@email', is_staff=True
+            username='change', password='change', email='email@email.com', is_staff=True
         )
         group = Group.objects.filter(name='Administrator')
         user.groups.set(group)
@@ -1273,7 +1387,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             user1 = self._create_user(
                 username='change1',
                 password='change1',
-                email='email1@email',
+                email='email1@email.com',
                 is_staff=True,
             )
             user1.groups.set(group)
@@ -1330,7 +1444,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
                 'post': 'yes',
             }
             url = reverse(f'admin:{self.app_label}_organizationuser_changelist')
-            with self.assertNumQueries(8):
+            # django-reversion adds ~4 queries
+            with self.assertNumQueries(12):
                 r = self.client.post(url, post_data, follow=True)
             qs = OrganizationUser.objects.filter(user=user1, organization=org1)
             self.assertEqual(r.status_code, 200)
@@ -1353,7 +1468,8 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             )
             self.assertContains(r, msg)
             post_data.update({'post': 'yes'})
-            with self.assertNumQueries(17):
+            # django-reversion adds ~4 queries
+            with self.assertNumQueries(21):
                 r = self.client.post(url, post_data, follow=True)
             qs = OrganizationUser.objects.filter(pk__in=[org_user.pk, org_user2.pk])
             self.assertEqual(r.status_code, 200)
@@ -1361,8 +1477,7 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             self.assertEqual(qs.count(), 1)
             self.assertEqual(qs.first().organization, org1)
 
-    @patch('sys.stdout', devnull)
-    @patch('sys.stderr', devnull)
+    @capture_any_output()
     def test_admin_add_user_with_invalid_email(self):
         admin = self._create_admin()
         self.client.force_login(admin)
@@ -1381,10 +1496,29 @@ class TestUsersAdmin(TestOrganizationMixin, TestUserAdditionalFieldsMixin, TestC
             self.client.post(reverse(f'admin:{self.app_label}_user_add'), params)
             mocked.assert_called_once()
 
-    @classmethod
-    def tearDownClass(cls):
-        super().tearDownClass()
-        devnull.close()
+    def test_admin_menu_groups(self):
+        # Test menu group (openwisp-utils menu group) for User, Organization,
+        # Organization Owner and Organization User models
+        admin = self._create_admin()
+        self.client.force_login(admin)
+        models = [
+            'user',
+            'organization',
+            'organizationowner',
+            'organizationuser',
+            'group',
+        ]
+        response = self.client.get(reverse('admin:index'))
+        for model in models:
+            with self.subTest(f'test menu group link for {model} model'):
+                url = reverse(f'admin:{self.app_label}_{model}_changelist')
+                self.assertContains(response, f'class="mg-link" href="{url}"')
+        with self.subTest('test user and organization group is registered'):
+            self.assertContains(
+                response,
+                '<div class="mg-dropdown-label">Users & Organizations </div>',
+                html=True,
+            )
 
 
 class TestBasicUsersIntegration(
@@ -1432,18 +1566,26 @@ class TestBasicUsersIntegration(
         params = user.__dict__
         params['bio'] = 'Test change'
         params.pop('phone_number')
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
+        params['birth_date'] = user.date_joined.date()
         params = self._additional_params_pop(params)
         params.update(self._get_user_edit_form_inline_params(user, org))
+        url = reverse(f'admin:{self.app_label}_user_change', args=[user.pk])
         response = self.client.post(
-            reverse(f'admin:{self.app_label}_user_change', args=[user.pk]),
+            url,
             params,
             follow=True,
         )
         self.assertNotContains(response, 'Please correct the error below.')
         user.refresh_from_db()
         self.assertEqual(user.bio, params['bio'])
+        self.assertEqual(user.birth_date, params['birth_date'])
+        with self.subTest('test presence of fields'):
+            response = self.client.get(url)
+            self.assertContains(response, 'id_birth_date')
+            self.assertContains(response, 'notes for internal usage')
 
     def _delete_inline_org_user(self, is_admin=False):
         admin = self._create_admin()
@@ -1453,7 +1595,8 @@ class TestBasicUsersIntegration(
         self.client.force_login(admin)
         params = user.__dict__
         params.pop('phone_number')
-        params.pop('_password')
+        params.pop('password', None)
+        params.pop('_password', None)
         params.pop('last_login')
         params = self._additional_params_pop(params)
         params.update(self._get_user_edit_form_inline_params(user, org))
@@ -1479,6 +1622,18 @@ class TestBasicUsersIntegration(
     def test_delete_inline_owner_org_user(self):
         self._delete_inline_org_user(is_admin=True)
 
+    def test_login_page(self):
+        r = self.client.get(reverse('admin:login'))
+
+        with self.subTest('Test forgot password link'):
+            self.assertContains(
+                r, '<a href="/accounts/password/reset/">Forgot Password?</a'
+            )
+
+        with self.subTest('Test username label'):
+            self.assertContains(r, '<label class="required" for="id_username">')
+            self.assertContains(r, 'Email, phone number or username:')
+
 
 class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, TestCase):
     app_label = 'openwisp_users'
@@ -1511,8 +1666,15 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
         organization_user3 = self._create_org_user(
             organization=org3, user=operator, is_admin=True
         )
+        administrator = self._create_administrator()
+        organization_user4 = self._create_org_user(
+            organization=org3, user=administrator, is_admin=True
+        )
         organization_user31 = self._create_org_user(organization=org3, user=user3)
         organization_user1o = self._create_org_user(organization=org1, user=operator)
+        organization_user1a = self._create_org_user(
+            organization=org1, user=administrator
+        )
         data = dict(
             org1=org1,
             org2=org2,
@@ -1528,11 +1690,14 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
             organization_user12=organization_user12,
             organization_user22=organization_user22,
             organization_user3=organization_user3,
+            organization_user4=organization_user4,
             organization_user1o=organization_user1o,
+            organization_user1a=organization_user1a,
             organization_user31=organization_user31,
             organization_owner1=organization_owner1,
             organization_owner2=organization_owner2,
             operator=operator,
+            administrator=administrator,
         )
         return data
 
@@ -1543,7 +1708,7 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
 
     def test_multitenancy_organization_user_queryset(self):
         data = self._create_multitenancy_test_env()
-        self._make_org_manager(data['operator'], data['org1'])
+        self._make_org_manager(data['administrator'], data['org1'])
         self._test_multitenant_admin(
             url=reverse(f'admin:{self.app_label}_organizationuser_changelist'),
             hidden=[
@@ -1556,15 +1721,17 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
                 data['organization_user1o'].user.username,
                 data['organization_user3'].user.username,
             ],
+            administrator=True,
         )
 
     def test_multitenancy_organization_owner_queryset(self):
         data = self._create_multitenancy_test_env()
-        self._make_org_manager(data['operator'], data['org1'])
+        self._make_org_manager(data['administrator'], data['org1'])
         self._test_multitenant_admin(
             url=reverse(f'admin:{self.app_label}_organizationowner_changelist'),
             hidden=[data['organization_owner2'].organization_user.user.username],
             visible=[data['organization_owner1'].organization_user.user.username],
+            administrator=True,
         )
 
     def test_useradmin_specific_multitenancy_costraints(self):
@@ -1573,6 +1740,7 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
             url=reverse(f'admin:{self.app_label}_user_changelist'),
             visible=[data['user3'], data['operator']],
             hidden=[data['user2'], data['user22'], data['user1'], data['user12']],
+            administrator=True,
         )
 
     def test_multitenant_admin_manager_only(self):
@@ -1585,7 +1753,7 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
         staff.groups.add(admin_group)
         self._create_org_user(organization=other_org, user=staff, is_admin=False)
         self._create_org_user(organization=staff_org, user=staff, is_admin=True)
-        self._login(staff)
+        self._login(staff.username)
         user1 = self._create_user(username='user1__otherorg', email='user1@user1.org')
         self._create_org_user(organization=other_org, user=user1, is_admin=False)
         self._test_multitenant_admin(
@@ -1607,3 +1775,34 @@ class TestMultitenantAdmin(TestMultitenantAdminMixin, TestOrganizationMixin, Tes
             self.assertNotIn('test', owner_admin.multitenant_shared_relations)
         with self.subTest('original attribute unaffected'):
             self.assertIsNone(MultitenantAdminMixin.multitenant_shared_relations)
+
+    def test_organization_user_filter(self):
+        data = self._create_multitenancy_test_env()
+        self._make_org_manager(data['administrator'], data['org1'])
+        url = reverse(f'admin:{self.app_label}_user_changelist')
+
+        with self.subTest('test superadmin'):
+            user = User.objects.filter(is_superuser=True, is_staff=True).first()
+            self.client.force_login(user)
+            response = self.client.get(url)
+            for org in Organization.objects.all():
+                self.assertContains(response, f'href="?organization={org.pk}">')
+
+        with self.subTest('test non superadmin'):
+            user = User.objects.get(username='administrator')
+            self.client.force_login(user)
+            response = self.client.get(url)
+            for pk in user.organizations_managed:
+                self.assertContains(response, f'href="?organization={pk}">')
+            for org in Organization.objects.exclude(pk__in=user.organizations_managed):
+                self.assertNotContains(response, f'href="?organization={org.pk}">')
+
+        with self.subTest('test only 1 org'):
+            OrganizationUser.objects.filter(
+                organization__slug__in=['organization2', 'organization3']
+            ).delete()
+            del user.organizations_managed
+            self.assertEqual(len(user.organizations_managed), 1)
+            response = self.client.get(url)
+            for pk in user.organizations_managed:
+                self.assertNotContains(response, f'href="?organization={pk}">')
